use std::error::Error;
use std::io::{ErrorKind, Read, Write};
use std::net::{Shutdown, TcpListener, TcpStream};
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::{Arc, Mutex};
use std::time::Duration;
use std::{env, thread};

use rcgen::generate_simple_self_signed;
use rustls::crypto::aws_lc_rs::sign::any_supported_type;
use rustls::pki_types::PrivateKeyDer;
use rustls::pki_types::pem::PemObject;
use rustls::server::ResolvesServerCertUsingSni;
use rustls::sign::CertifiedKey;
use rustls::{ServerConfig, ServerConnection, StreamOwned};
use threadpool::ThreadPool;

fn accept_client(
    stream: Arc<TcpStream>,
    config: Arc<ServerConfig>,
    sni_domain: String,
    remote_stream: Arc<TcpStream>,
    threadpool: ThreadPool
) -> Result<(), Box<dyn Error>> {
    stream.set_nonblocking(true)?;
    remote_stream.set_nonblocking(true)?;

    let mut tls = StreamOwned::new(ServerConnection::new(config)?, stream.try_clone()?);

    while tls.conn.is_handshaking() {
        match tls.conn.complete_io(&mut tls.sock) {
            Ok(_) => {},
            Err(e) => {
                if e.kind() != ErrorKind::WouldBlock {
                    return Err(e.into());
                }
            }
        }
    }

    if tls.conn.server_name() != Some(&sni_domain) {
        return Err("unexpected sni domain".into());
    }

    let break_it = Arc::new(AtomicBool::new(false));

    let tls = Arc::new(Mutex::new(tls));

    threadpool.execute({
        let tls = tls.clone();
        let remote_stream = remote_stream.clone();
        let break_it = break_it.clone();

        move || {
            while !break_it.load(Ordering::SeqCst) {
                let mut buffer = vec![0; 4096];
                match (&*remote_stream).read(&mut buffer) {
                    Ok(0) => { break },
                    Ok(n) => {
                        // println!("from remote {n}");
                        buffer.truncate(n);
                        match tls.lock().unwrap().write(&buffer) {
                            Ok(_) => {},
                            Err(e) => {
                                if e.kind() != ErrorKind::WouldBlock {
                                    break;
                                }
                            }
                        }
                    }
                    Err(e) => {
                        if e.kind() != ErrorKind::WouldBlock {
                            break;
                        }
                    }
                }
                thread::sleep(Duration::from_millis(2));
            }

            break_it.store(true, Ordering::SeqCst);
        }
    });

    let tls = tls.clone();
    let remote_stream = remote_stream.clone();

    while !break_it.load(Ordering::SeqCst) {
        let mut buffer = vec![0; 4096];
        match tls.lock().unwrap().read(&mut buffer) {
            Ok(0) => { 
                break_it.store(true, Ordering::SeqCst);
                break; 
            },
            Ok(n) => {
                // println!("from tls {n}");
                buffer.truncate(n);
                match (&*remote_stream).write(&buffer) {
                    Ok(_) => {},
                    Err(e) => {
                        if e.kind() != ErrorKind::WouldBlock {
                            break_it.store(true, Ordering::SeqCst);
                            return Err(e.into());
                        }
                    }
                }
            }
            Err(e) => {
                if e.kind() != ErrorKind::WouldBlock {
                    break_it.store(true, Ordering::SeqCst);
                    return Err(e.into());
                }
            }
        }
        thread::sleep(Duration::from_millis(2));
    }

    Ok(())
}

fn main() -> Result<(), Box<dyn Error>> {
    let mut args = env::args();

    args.next();

    let sni_domain = args.next().expect("missing sni domain argument");
    let host = args.next().expect("missing host argument");
    let local_host = args.next().expect("missing local host argument");

    let certified_key = generate_simple_self_signed(vec![sni_domain.to_string()]).unwrap();
    let certified_key = CertifiedKey::new(
        vec![certified_key.cert.into()],
        any_supported_type(&PrivateKeyDer::from_pem_slice(
            certified_key.key_pair.serialize_pem().as_bytes(),
        )?)?,
    );

    let mut resolver = ResolvesServerCertUsingSni::new();
    resolver.add(&sni_domain, certified_key)?;

    let config = Arc::new(
        ServerConfig::builder()
            .with_no_client_auth()
            .with_cert_resolver(Arc::new(resolver)),
    );

    let threadpool = ThreadPool::new(100);

    let listener = TcpListener::bind(local_host).unwrap();

    for stream in listener.incoming() {
        let stream = stream.expect("listener got broken");

        threadpool.execute({
            let config = config.clone();
            let sni_domain = sni_domain.clone();
            let host = host.clone();
            let threadpool = threadpool.clone();

            move || {
                let Ok(remote_stream) = TcpStream::connect(host) else {
                    return;
                };

                let addr = stream.peer_addr().unwrap();

                println!("connected {}", addr);

                let stream = Arc::new(stream);
                let remote_stream = Arc::new(remote_stream);

                if let Err(e) = accept_client(stream.clone(), config, sni_domain, remote_stream, threadpool) {
                    println!("error connection: {e:?}");
                }

                let _ = (&*stream).shutdown(Shutdown::Both);

                println!("disconnected {}", addr);
            }
        });
    }

    Ok(())
}
